/*
 * Copyright (c) 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <MapUnidraw/mapcomps.h>
#include <UtmUnidraw/utmfunc.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/maproute.h>
#include <MapUnidraw/mapviewstate.h>
#include <Map/mapelt.h>
#include <OverlayUnidraw/ovcomps.h>
#include <OverlayUnidraw/ovselection.h>
#include <OverlayUnidraw/ovviewer.h>
#include <OverlayUnidraw/ovviews.h>
#include <Unidraw/Commands/edit.h>
#include <Unidraw/Commands/transforms.h>
#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Components/component.h>
#include <Unidraw/Components/compview.h>
#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/selection.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>
#include <ComTerp/comvalue.h>
#include <Attribute/aliterator.h>
#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/paramlist.h>
#include <TopoFace/fgeomobjs.h>
#include <InterViews/canvas.h>
#include <InterViews/transformer.h>
#include <ctype.h>
#include <strstream.h>
#include <math.h>

/*****************************************************************************/

UtmFunc::UtmFunc(ComTerp* comterp, Editor* ed) : UnidrawFunc(comterp, ed) {
}

float UtmFunc::heading(float dx, float dy) {
  if (dx < 0) {
      if (dy < 0) 
	  return 180.0 + atan(-dx/-dy) * RAD_TO_DEG;
      else if (dy == 0) 
	  return 270.0;
      else /* dy > 0 */
	  return 270.0 + atan(dy/-dx) * RAD_TO_DEG;
  } else if (dx == 0) {
      if (dy < 0) 
          return 180.0;
      else /* dy >= 0 */
	  return 0.0;
  } else /* dx > 0 */ {
      if (dy < 0) 
	  return 90.0 + atan(-dy/dx) * RAD_TO_DEG;
      else if (dy == 0) 
	  return 90.0;
      else 
	  return atan(dx/dy) * RAD_TO_DEG;
  }
}

/*****************************************************************************/

VhclViewFunc::VhclViewFunc(ComTerp* comterp, Editor* ed) : UtmFunc(comterp, ed) {
    _name_symid = symbol_add("name");
    _gr_symid = symbol_add("gr");
    _vhclpict_symid = symbol_add("VhclPict");
}

void VhclViewFunc::execute() {
    ComValue grstr(stack_key(_gr_symid));
    ComValue namestr(stack_key(_name_symid));
    reset_stack();

    /* kludge to globally disable handles mechanism */
    ((OverlaySelection*)_ed->GetSelection())->DisableHandles();

    istrstream in(grstr.string_ptr());
    const char* first_token = "drawtool";
    int len = strlen(first_token)+1;
    char buf[len];

    char ch;
    while (isspace(ch = in.get())); in.putback(ch);
    ParamList::parse_token(in, buf, len);
    if (strcmp(buf, first_token)!=0) {
        push_stack(ComValue::nullval());
	return;
	}
    
    OverlaysComp* vhclpict = new OverlaysComp(in, nil);
    if (vhclpict->valid()) {

        AttributeList* alist = vhclpict->GetAttributeList();
	alist->add_attr("name", namestr);

	PasteCmd* cmd = new PasteCmd(_ed, new Clipboard(vhclpict));
	execute_log(cmd);

        ComValue result(ComValue::ObjectType, (void*)new ComponentView(vhclpict));
        result.obj_type_ref() = _vhclpict_symid;
        push_stack(result);

    } else 
        push_stack(ComValue::nullval());
}

/*****************************************************************************/
 
VhclMoveToFunc::VhclMoveToFunc(ComTerp* comterp, Editor* ed) : UtmFunc(comterp, ed) {
    _heading_symid = symbol_add("heading");
}

void VhclMoveToFunc::execute() {
    ComValue compsym(stack_arg(0, true));
    ComValue compval(lookup_symval(compsym));
    ComValue utm_xloc(stack_arg(1));
    ComValue utm_yloc(stack_arg(2));
    ComValue heading(stack_key(_heading_symid));
    reset_stack();

    float xloc, yloc;
    Transformer* trans = ((MapEditor*)_ed)->GetMapViewState()->deftrans();
    if (trans)
      trans->transform(utm_xloc.float_val(), utm_yloc.float_val(),
		       xloc, yloc);
    else {
      xloc = utm_xloc.float_val();      
      yloc = utm_yloc.float_val();
    }

    ComponentView* compview = 
        (ComponentView*)compval.geta(symbol_add("VhclPict"));
    OverlaysComp* vhclpict = 
       compview ? (OverlaysComp*)compview->GetSubject() : nil;
    if (vhclpict) {
        Selection* sel = _ed->GetSelection();
	sel->Clear();
	OverlayView* view = vhclpict->FindView(_ed->GetViewer());
	if (view) {
	    sel->Append(view);
	    Transformer * trans = vhclpict->GetGraphic()->GetTransformer();
	    float dx, dy;
	    if (trans) {
	      float a00, a01, a10, a11, a20, a21;
	      trans->matrix(a00, a01, a10, a11, a20, a21);
	      Transformer newt(a00, a01, a10, a11, 0.0, 0.0);
	      newt.transform(0.0, 1.0, dx, dy);
	    } else {
	      dx = 0;
	      dy = 1;
	    }
	    float old_heading = UtmFunc::heading(dx, dy);
	    float new_heading = heading.float_val();
	    MacroCmd* mcmd = new MacroCmd(_ed);
	    if (old_heading != new_heading) {
	        float rotation = old_heading-new_heading;
	        RotateCmd* cmd = new RotateCmd(_ed, rotation);
		mcmd->Append(cmd);
	    }
	    float cx, cy;
	    vhclpict->GetGraphic()->GetCenter(cx, cy);
            MoveCmd* cmd = new MoveCmd(_ed, xloc-cx, yloc-cy);
	    mcmd->Append(cmd);
	    execute_log(mcmd);

	    // recenter view on vehicle if it moves out of view
	    OverlayViewer* viewer = (OverlayViewer*)_ed->GetViewer();
	    Canvas* canvas = viewer->GetCanvas();
	    int pw = canvas->pwidth();
	    int ph = canvas->pheight();
	    Transformer* rel = viewer->GetRel();
	    Transformer* utrans = ((MapEditor*)_ed)->GetMapViewState()->deftrans();
	    Transformer ident;
	    if (!utrans) utrans = &ident;
	    float rx0, ry0, rx1, ry1, ux0, uy0, ux1, uy1;
	    rel->inverse_transform(10.0, 10.0, rx0, ry0);
	    utrans->inverse_transform(rx0, ry0, ux0, uy0);
	    rel->inverse_transform(float(pw-101), float(ph-101), rx1, ry1);
	    utrans->inverse_transform(rx1, ry1, ux1, uy1);
	    FBoxObj utmviewbox(ux0, uy0, ux1, uy1);
	    FPointObj utmpoint(utm_xloc.float_val(), utm_yloc.float_val());
	    if (!utmviewbox.Contains(utmpoint)) {
	      float csx = (pw-1)/2.0;
	      float csy = (ph-1)/2.0;
	      float vrx, vry, vsx, vsy;
	      utrans->transform(utm_xloc.float_val(), utm_yloc.float_val(), 
				vrx, vry);
	      rel->transform(vrx, vry, vsx, vsy);
	      int px = int(vsx - csx);
	      int py = int(vsy - csy);
	      PanCmd* pancmd = new PanCmd(_ed, px, py);
	      execute_log(pancmd);
	    }
	}
        push_stack(compval);
    } else 
        push_stack(ComValue::nullval());
    return;
}


/*****************************************************************************/

RteViewFunc::RteViewFunc(ComTerp* comterp, Editor* ed) : UtmFunc(comterp, ed) {
    _name_symid = symbol_add("name");
    _rtepict_symid = symbol_add("RtePict");
}

void RteViewFunc::execute() {
    ComValue points(stack_arg(0));
    ComValue* namestr = new ComValue(stack_key(_name_symid));
    reset_stack();

    if (!points.is_type(ComValue::ArrayType)) {
       push_stack(ComValue::nullval());
       return;
    }

    int npts = points.array_len()/2;
    double* x = new double[npts];
    double* y = new double[npts];
    ALIterator i;
    AttributeValueList* al = points.array_val();
    al->First(i);
    for (int j=0; j<npts; j++) {
        x[j] = al->GetAttrVal(i)->double_val();
        al->Next(i);
        y[j] = al->GetAttrVal(i)->double_val();
        al->Next(i);
    }

    MapElement* route = new MapElement(0, MapFeature::LineType,
				       npts, x, y);
    MapRouteComp* rtepict = new MapRouteComp(route);
    if (rtepict->valid()) {

        AttributeList* alist = rtepict->GetAttributeList();
	alist->add_attr("name", namestr);

	PasteCmd* cmd = new PasteCmd(_ed, new Clipboard(rtepict));
	execute_log(cmd);

        ComValue result(ComValue::ObjectType, (void*)new ComponentView(rtepict));
        result.obj_type_ref() = _rtepict_symid;
        push_stack(result);

    } else {
      delete namestr;
      push_stack(ComValue::nullval());
    }
}

/*****************************************************************************/

MapViewStateFunc::MapViewStateFunc(ComTerp* comterp, Editor* ed) 
: UtmFunc(comterp, ed) {
}

void MapViewStateFunc::execute() {

    static int deftrans_symid = symbol_add("deftrans");
    ComValue deftransv(stack_key(deftrans_symid));
    static int normtrans_symid = symbol_add("normtrans");
    ComValue normtransv(stack_key(normtrans_symid));
    reset_stack();

    if (deftransv.is_array() && deftransv.array_len()==6) {

      AttributeValueList* avl = deftransv.array_val();
      Iterator i;
      avl->First(i);
      float af[6];
      for (int j=0; j<6; j++) {
	af[j] = avl->GetAttrVal(i)->float_val();
	avl->Next(i);
      }

      MapEditor* ed = (MapEditor*)_ed;
      MapViewState* mapviewstate = ed->GetMapViewState();
      mapviewstate->settrans(new Transformer(af[0], af[1], af[2], af[3], af[4], af[5]));
      mapviewstate->normtrans(normtransv.boolean_val());
    }
}

